using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using static SOMDocManager;

public class CanonBallProblemCalculator
{
    private int Dim_const, Dim_A, Dim_B;
    private int MaxIter;

    private FactOrganizer FactOrganizer;

    private Vector3 StartPos;
    private Vector3 StartVel;
    private Vector3 Gravity;

    private List<Wall> Walls;
    public string Result_MovementFunc_Id;
    public List<string> Result_ArgsFunc_Id;
    public List<string> Result_FuncCall_Id;

    public struct Wall
    {
        public LineFact Topology;
        public float Bounce;

        public Wall(LineFact Topology, float Bounce)
        {
            this.Bounce = Bounce;
            this.Topology = Topology;
        }
    }

    public CanonBallProblemCalculator(List<Wall> walls, Vector3 starPos, Vector3 starVec, Vector3 gravity, int dim_const, int dim_A, int dim_B, FactOrganizer factOrganizer, int maxIter = 256)
    {
        Walls = walls;
        Dim_const = dim_const;
        Dim_A = dim_A;
        Dim_B = dim_B;
        MaxIter = maxIter;
        StartPos = starPos;
        StartVel = starVec;
        Gravity = gravity;
        FactOrganizer = factOrganizer;

        Compute();
    }

    private void Compute()
    {
        SOMDoc BuildOMA_XVA()
        {   //Pos = Pos + Vel * t + 0.5 * g * t**2
            return 
            new OMBINDC("Pos", MMT_OMS_URI.Tuple,
            new OMBINDC("Vel", typeof(Vector3),
            new OMBINDC("Acc", typeof(Vector3),
            new OMBINDC("t", "R",
                new OMA(
                new OMS(MMT_OMS_URI.Add),
                new() {
                        new OMV("Pos"),
                        new OMA(
                            new OMS(MMT_OMS_URI.Add),
                            new() {
                                new OMA(
                                    new OMS(MMT_OMS_URI.Multiply),
                                    new() {
                                        new OMV("Vel"),
                                        new OMV("t"),
                                    }
                                ),
                                new OMA(
                                    new OMS(MMT_OMS_URI.Multiply),
                                    new() {
                                        new OMF(0.5f),
                                        new OMA(
                                            new OMS(MMT_OMS_URI.Multiply),
                                            new() {
                                                new OMV("Acc"),
                                                new OMA(
                                                    new OMS(MMT_OMS_URI.Multiply),
                                                    new() {
                                                        new OMV("t"),
                                                        new OMV("t"),
            }),}),}),}),})))));
        }
        
        SOMDoc BuildOMAPath(Vector3 Pos, Vector3 Vel)
        {   //t -> [Pos, Vel, g, t]
            return new OMBINDC("t", "R",
                new OMA(
                    new OMS(MMT_OMS_URI.MakeObjectArray),
                    new() {
                        SOMDoc.MakeUnityEngineVector3(Pos),
                        SOMDoc.MakeUnityEngineVector3(Vel),
                        SOMDoc.MakeUnityEngineVector3(Gravity),
                        new OMV("t"),
                    }
            ));
        }

        //Pos = Pos + Vel * t + 0.5 * g * t**2
        Vector3 UpdatePos(Vector3 Pos, Vector3 Vel, float t)
            => Pos + Vel * t + 0.5f * t * t * Gravity;

        //Vel = Vel + g * t
        Vector3 UpdateVel(Vector3 Vel, float t)
            => Vel + Gravity * t;

        // p1 + v1 * t + 0.5 * g1 * t * t = o1 + d1 * x,
        // p2 + v2 * t + 0.5 * g2 * t * t = o2 + d2 * x,

        // T = (-b  sqrt(b ^ 2 - 4ac)) / (2a)
        //    Where:
        //        a = 0.5 * (G1 * D2 - G2 * D1)
        //        b = V1 * D2 - V2 * D1
        //        c = - P2 * D1 + P1 * D2 + O2 * D1 - O1 * D2
        (float, float) SolveForTime(Vector3 Pos, Vector3 Vel, LineFact Top)
        {
            double
                a = 0.5 * (Gravity[Dim_A] * Top.Dir[Dim_B] - Gravity[Dim_B] * Top.Dir[Dim_A]),
                b = Vel[Dim_A] * Top.Dir[Dim_B] - Vel[Dim_B] * Top.Dir[Dim_A],
                c = - Pos[Dim_B] * Top.Dir[Dim_A] + Pos[Dim_A] * Top.Dir[Dim_B]
                    + Top.Point1.Point[Dim_B] * Top.Dir[Dim_A] - Top.Point1.Point[Dim_A] * Top.Dir[Dim_B];

            if(a == 0)
                return ((float) (-c/b), -1);

            float
                t1 = (float) ((-b + Math.Sqrt(b * b - 4 * a * c)) / (2 * a)),
                t2 = (float) ((-b - Math.Sqrt(b * b - 4 * a * c)) / (2 * a));

            return (t1, t2);
        };

        // X = (P1 + V1 * T + 0.5 * G1 * T ^ 2 - O1) / D1
        float SolveForDistanceOnPlane(Vector3 Pos, Vector3 Vel, LineFact Top, float t)
        {
            Vector3 dir = Top.Point1.Point - UpdatePos(Pos, Vel, t);

            float dot = Vector3.Dot(dir, Top.Dir);
            if (float.IsNaN(dot))
                return -1;

            return dir.magnitude * Math.Sign(dot);
        };

        var Result_Movement = new FunctionFact(BuildOMA_XVA(), FactOrganizer);
        Result_MovementFunc_Id = FactOrganizer.Add(Result_Movement, out _, false, null, null);

        Vector3 PlaneNorm = Vector3.zero;
        PlaneNorm[Dim_const] = 1;

        Vector3 pos = StartPos;
        Vector3 vel = StartVel;

        Result_ArgsFunc_Id = new();
        Result_FuncCall_Id = new();

        for (int i = 0; i < MaxIter; i++)
        {
            pos = UpdatePos(pos, vel, (float) Math3d.vectorPrecission); //minimum step

            var hits = Walls
                .Select(w => (w, tarr: SolveForTime(pos, vel, w.Topology)))
                .SelectMany(wt => new[] { (wt.w, t: wt.tarr.Item1), (wt.w, t: wt.tarr.Item2) })
                .Where(wt => wt.t >= 0)
                .Select(wt => (wt.w, wt.t, x: SolveForDistanceOnPlane(pos, vel, wt.w.Topology, wt.t)))
                .Where(wt => wt.x >= 0 && wt.x <= wt.w.Topology.Distance)
                .OrderBy(wt => wt.t)
                .ToArray();

            (Wall next_Wall, float next_t, float x_fac_debug) = hits.Length > 0 
                ? hits[0] 
                : (default(Wall), float.PositiveInfinity, default);

            Result_ArgsFunc_Id.Add(
                FactOrganizer.Add(
                    new FunctionFact(BuildOMAPath(pos, vel), FactOrganizer),
                    out _, true, null, null));

            Result_FuncCall_Id.Add(
                FactOrganizer.Add(
                    new FunctionCallFact(Result_MovementFunc_Id, Result_ArgsFunc_Id.Last(), (-(float)Math3d.vectorPrecission, next_t), FactOrganizer),
                    out _, true, null, null));

            if (hits.Length == 0)
                break;

            pos = UpdatePos(pos, vel, next_t);
            Vector3 vel_at_bounce = UpdateVel(vel, next_t);

            Vector3 LineFactNorm = Vector3.Cross(PlaneNorm, next_Wall.Topology.Dir);
            float bounce_angle_cos = Mathf.Abs(Vector3.Dot(vel_at_bounce.normalized, LineFactNorm));
            float bounce_fac = 1 - (1 - next_Wall.Bounce) * (Mathf.Acos(bounce_angle_cos) / 2 / Mathf.PI);

            vel = bounce_fac * Vector3.Reflect(vel_at_bounce, LineFactNorm);
        }

        return;
    }
}
